#! /usr/bin/env racket

#lang racket
(require racket/include)
(require parser-tools/yacc
         parser-tools/lex
         (prefix-in : parser-tools/lex-sre))

(require "read-command-line.rkt")
(require (prefix-in line-counter: "line-counter.rkt"))
(require "configurations.rkt")
(require (prefix-in errors: "errors.rkt"))
(require (prefix-in schemata-manager: "schemata-manager.rkt"))
(require (prefix-in parser: "parser.rkt"))
(require "preprocessor.rkt")

;this stores the closure that manage all the gmql parsing
(define gmql-parser  (λ args (error "gmql-parser has not been initialized yet")))

;this is the new input-port with all the comments removed
(define uncommented-input-port 'null)

;stores a closure that contains all the schemata that
; are used in the GMQL program
(define global-schemata (λ args (error "global-schemata has not been initialized yet")))

(with-handlers
  ([exn:fail? errors:exception-fail-handler])
   (let
     
     ([gmql-input-file (open-input-file input-file-path)])
     
     (begin
       
       ;set global schemata
       (set! global-schemata (schemata-manager:get-new-schemata))
       (global-schemata 'fill allowed-users)

       (set! uncommented-input-port (remove-comments (open-input-file input-file-path)))
       
       ;set parser
       (set! gmql-parser (parser:make-parser global-schemata))
       (gmql-parser 'parse-input-port uncommented-input-port)
       
       ;output the pig code
       (with-output-to-file generated-pig-code-file-path
			    (λ ()
			      (begin
				(displayln (gmql-parser 'get-preamble))
				(displayln (gmql-parser 'get-pig-code))))
			    #:mode 'text
			    #:exists 'replace)
       
       ;output the materialize schemata
       (with-output-to-file generated-schema-output-file
			    (λ ()
				(displayln (gmql-parser 'get-schemata)))
			    #:mode 'text
			    #:exists 'replace)

       )))

