#lang racket

(require parser-tools/yacc
	 parser-tools/lex
	 (prefix-in : parser-tools/lex-sre)
	 "data-structures.rkt"
	 "configurations.rkt"
	 (prefix-in line-counter: "line-counter.rkt")
	 (prefix-in parser-utils: "parser-utils.rkt")
	 (prefix-in schemata-manager: "schemata-manager.rkt"))


(provide make-parser)

(define-tokens value-tokens   ;;token which have a value
	       (STRING-VALUE 
		 INT-VALUE 
		 FLOAT-VALUE
		 VAR-ID-VALUE))

(define-tokens function-tokens
	       (BUILTIN-NULLARY-FUNCTION-NAME
		BUILTIN-UNARY-FUNCTION-NAME))

(define-empty-tokens 
  op-tokens ;;token without a values
  (
   SELECT 
   PROJECT 
   MAP
   MAP-STRANDED
   JOIN
   JOIN-STRANDED
   UNION
   COVER
   COVER-FLAT
   SUMMIT
   SUMMIT-FLAT
   DIFFERENCE
   DIFFERENCE-STRANDED
   MATERIALIZE
   ORDER
   TOP
   TOPG
   DESC
   MERGE
   JACCARD
   GROUP
   BY
   ANY
   ALL
   LEFT
   RIGHT
   LEFT-DISTINCT
   RIGHT-DISTINCT
   PROJECT-LEFT
   PROJECT-RIGHT
   PROJECT-LEFT-DISTINCT
   PROJECT-RIGHT-DISTINCT
   PLUS
   MINUS
   STAR
   DIV 
   ASSIGN
   AS
   EQ 
   NOT-EQ
   LOWER 
   LOWER-EQ
   GREATER 
   GREATER-EQ
   INTERSECTION 
   CONJUNCTION 
   L-PAR 
   R-PAR 
   R-ARROW
   OR 
   AND
   NOT
   TRUE
   DISTANCE 
   MIN-DISTANCE
   FIRST
   AFTER
   UPSTREAM-DISTANCE
   DOWNSTREAM-DISTANCE
   AGGREGATE
   START
   STOP
   STD
   MEDIAN
   CHR
   STRAND
   SEMI-COLON 
   COMMA 
   DOT
   EOF
   ))

(define-lex-abbrevs ;;abbreviation (i.e. regular expressions)
  [ASSIGN "="]
  [EQ "=="]
  [NOT-EQ "!="]
  [R-ARROW "->"]
  [LOWER "<"]
  [LOWER-EQ "<="]
  [GREATER ">"]
  [GREATER-EQ ">="]
  [SEMI-COLON ";"]
  [COMMA ","]
  [DOT "."]
  [OR (:+ "OR" "or")]
  [AND (:+ "AND" "and")]
  [NOT (:+ "NOT" "not")]
  [TRUE (:+ "TRUE" "true" "True")]
  [L-PAR "("]
  [R-PAR ")"]
  [INT (:+ (:/ "0" "9"))]
  [FLOAT (:seq (:+ (:/ "0" "9")) (:seq ".") (:+ (:/ "0" "9")))]
  [SELECT (:+ "SELECT" "select")]
  [PROJECT (:+ "PROJECT" "project")]
  [MAP (:+ "MAP" "map")]
  [MAP-STRANDED (:+ "MAP_STRANDED" "map_stranded")]
  [JOIN (:+ "JOIN" "join")]
  [JOIN-STRANDED (:+ "JOIN_STRANDED" "join_stranded")]
  [UNION (:+ "UNION" "union")]
  [COVER (:+ "COVER" "cover")]
  [COVER-FLAT (:+ "COVER_FLAT" "cover_flat")]
  [SUMMIT (:+ "SUMMIT" "summit")]
  [SUMMIT-FLAT (:+ "SUMMIT_FLAT" "summit_flat")]
  [DIFFERENCE (:+ "DIFFERENCE" "difference")]
  [DIFFERENCE-STRANDED (:+ "DIFFERENCE_STRANDED" "difference_stranded")]
  [ORDER (:+ "ORDER" "order")]
  [TOP (:+ "TOP" "top")]
  [TOPG (:+ "TOPG" "topg")]
  [DESC (:+ "DESC" "desc")]
  [GROUP (:+ "GROUP" "group")]
  [BY (:+ "BY" "by")]
  [MERGE (:+ "MERGE" "merge")]
  [JACCARD (:+ "JACCARD" "jaccard")]
  [ANY (:+ "ANY" "any")]
  [ALL (:+ "ALL" "all")]
  [MATERIALIZE (:+ "MATERIALIZE" "materialize")]
  [LEFT (:+ "left" "LEFT")]        ;cambiare in r1 e r2, o altro
  [RIGHT (:+ "right" "RIGHT")]
  [LEFT-DISTINCT (:+ "left_distinct" "LEFT_DISTINCT")]
  [RIGHT-DISTINCT (:+ "right_distinct" "RIGHT_DISTINCT")]
  [PROJECT-LEFT (:+ "project_left" "PROJECT_LEFT")]
  [PROJECT-RIGHT (:+ "project_right" "PROJECT_RIGHT")]
  [PROJECT-LEFT-DISTINCT (:+ "project_left_distinct" "PROJECT_LEFT_DISTINCT")]
  [PROJECT-RIGHT-DISTINCT (:+ "project_right_distinct" "PROJECT_RIGHT_DISTINCT")]
  [INTERSECTION (:or "INT" "intersection" "int" "INTERSECTION")]
  [CONJUNCTION (:+ "CAT" "cat")]
  [DISTANCE (:+ "DISTANCE" "distance")]
  [MIN-DISTANCE (:+ "MINDISTANCE" "mindistance")]
  [FIRST (:+ "FIRST" "first")]
  [AFTER (:+ "AFTER" "after")]
  [AS (:+ "AS" "as")]
  [UPSTREAM-DISTANCE (:+ "UPSTREAM_DISTANCE" "upstream_distance")]
  [DOWNSTREAM-DISTANCE (:+ "DOWNSTREAM_DISTANCE" "downstream_distance")]
  [AGGREGATE (:+ "AGGREGATE" "aggregate")]
  [START (:+ "start" "START")]
  [DIV (:+ "/")]
  [STOP (:+ "stop" "STOP")]
  [CHR (:+ "chr" "CHR")]
  [PLUS "+"]
  [MINUS "-"]
  [STAR "*"]
  [STRAND (:+ "strand" "STRAND")]
  [VAR-ID (:seq (:* "_") (:or (:/ "a" "z") (:/ "A" "Z") ) (:* (:or (:/ "a" "z") (:/ "A" "Z") (:/ "0" "9") "_")))]
  [STRING (:seq (:seq "'") (:+ (:or (:/ "a" "z") (:/ "A" "Z") (:/ "0" "9") "(" ")" "|" "!" "?" "[" "]" "{" "}" "," ";" ":" "." "_" "-" "*" "+" " ")) (:seq "'"))])


;get a parameter, the initial set of global schemata
(define (make-parser gs)
  (letrec
    ((counter (line-counter:get-new-counter))
     (global-schemata gs)
     (preamble "")
     (pig-code "")
     (output-schemata "")
     (lex-token (lexer ; the input of this function is input-port?
		  [(eof) 'EOF]
		  ;; recursively call the lexer on the remaining input after a tab or space.  Returning the "1+1")
		  ;; result of that operation.  This effectively skips all whitespace. 
		  [(:seq #\newline) (begin (counter '++) (lex-token input-port))]  
		  [(:or #\tab #\space)  (lex-token input-port)]
		  [(:+ INT) (token-INT-VALUE (string->number lexeme))]
		  [(:+ FLOAT) (token-FLOAT-VALUE (string->number lexeme))]
		  [(:seq AS) 'AS]
		  [(:seq SELECT) 'SELECT]
		  [(:seq PROJECT) 'PROJECT]
		  [(:seq MAP) 'MAP]
		  [(:seq MAP-STRANDED) 'MAP-STRANDED]
		  [(:seq JOIN) 'JOIN]
		  [(:seq JOIN-STRANDED) 'JOIN-STRANDED]
		  [(:seq UNION) 'UNION]
		  [(:seq MATERIALIZE) 'MATERIALIZE]
		  [(:seq COVER) 'COVER]
		  [(:seq COVER-FLAT) 'COVER-FLAT]
		  [(:seq SUMMIT) 'SUMMIT]
		  [(:seq SUMMIT-FLAT) 'SUMMIT-FLAT]
		  [(:seq DIFFERENCE) 'DIFFERENCE]
		  [(:seq DIFFERENCE-STRANDED) 'DIFFERENCE-STRANDED]
		  [(:seq MERGE) 'MERGE]
		  [(:seq JACCARD) 'JACCARD]
                  [(:seq ORDER) 'ORDER]
                  [(:seq DESC) 'DESC]
                  [(:seq TOP) 'TOP]
                  [(:seq TOPG) 'TOPG]
		  [(:seq GROUP) 'GROUP]
		  [(:seq BY) 'BY]
		  [(:seq ANY) 'ANY]
		  [(:seq ALL) 'ALL]
		  [(:seq LEFT) 'LEFT]
		  [(:seq RIGHT) 'RIGHT]
		  [(:seq LEFT-DISTINCT) 'LEFT-DISTINCT]
		  [(:seq RIGHT-DISTINCT) 'RIGHT-DISTINCT]
		  [(:seq PROJECT-LEFT) 'PROJECT-LEFT]
		  [(:seq PROJECT-RIGHT) 'PROJECT-RIGHT]
		  [(:seq PROJECT-LEFT-DISTINCT) 'PROJECT-LEFT-DISTINCT]
		  [(:seq PROJECT-RIGHT-DISTINCT) 'PROJECT-RIGHT-DISTINCT]
		  [(:seq AGGREGATE) 'AGGREGATE]
		  [(:seq EQ) 'EQ]
		  [(:seq NOT-EQ) 'NOT-EQ]
		  [(:seq LOWER) 'LOWER]
		  [(:seq LOWER-EQ) 'LOWER-EQ]
		  [(:seq R-ARROW) 'R-ARROW]
		  [(:seq GREATER) 'GREATER]
		  [(:seq GREATER-EQ) 'GREATER-EQ]
		  [(:seq ASSIGN) 'ASSIGN]
		  [(:seq L-PAR) 'L-PAR]
		  [(:seq R-PAR) 'R-PAR]
		  [(:seq OR) 'OR]
		  [(:seq AND) 'AND]
		  [(:seq NOT) 'NOT]
                  [(:seq TRUE) 'TRUE]
		  [(:seq SEMI-COLON) 'SEMI-COLON]
		  [(:seq COMMA) 'COMMA]
		  [(:seq DOT) 'DOT]
		  [(:seq "COUNT") (token-BUILTIN-NULLARY-FUNCTION-NAME lexeme)] 
		  [(:seq "EXISTS") (token-BUILTIN-NULLARY-FUNCTION-NAME lexeme)]
		  [(:seq "AVG") (token-BUILTIN-UNARY-FUNCTION-NAME lexeme)]
		  [(:seq "MIN") (token-BUILTIN-UNARY-FUNCTION-NAME lexeme)]
		  [(:seq "MAX") (token-BUILTIN-UNARY-FUNCTION-NAME lexeme)]
		  [(:seq "SUM") (token-BUILTIN-UNARY-FUNCTION-NAME lexeme)]
                  [(:seq "BAG") (token-BUILTIN-UNARY-FUNCTION-NAME lexeme)]
		  [(:seq "STD") (token-BUILTIN-UNARY-FUNCTION-NAME lexeme)]
                  [(:seq "MEDIAN") (token-BUILTIN-UNARY-FUNCTION-NAME lexeme)]
                  [(:seq INTERSECTION) 'INTERSECTION]
		  [(:seq CONJUNCTION) 'CONJUNCTION]
		  [(:seq DISTANCE) 'DISTANCE]
		  [(:seq MIN-DISTANCE) 'MIN-DISTANCE]
		  [(:seq FIRST) 'FIRST]
		  [(:seq AFTER) 'AFTER]
		  [(:seq UPSTREAM-DISTANCE) 'UPSTREAM-DISTANCE]
		  [(:seq DOWNSTREAM-DISTANCE) 'DOWNSTREAM-DISTANCE]
		  [(:seq START) 'START]
		  [(:seq STOP) 'STOP]
		  [(:seq CHR) 'CHR]
		  [(:seq STRAND) 'STRAND]
		  [(:seq PLUS) 'PLUS]
		  [(:seq MINUS) 'MINUS]
		  [(:seq STAR) 'STAR]
		  [(:seq VAR-ID) (token-VAR-ID-VALUE (list lexeme (counter 'get) ))]
		  [(:seq STRING) (token-STRING-VALUE (substring lexeme 1 (- (string-length lexeme) 1)))] 
		  [(:seq DIV) 'DIV]
                   ))

     (gql-parser (parser
		   (start query)
		   (end EOF)
		   (tokens value-tokens op-tokens function-tokens)
		   (suppress) ;; Our grammar is ambiguous. By default, parser complains.
		   (error (λ(ok? name value)
			    (error  (string-append 
				      "Unexpected token: '"  
				      (cond
					[(equal? value #f) 
                                         (cond
                                          [(equal? name 'L-PAR) "("]
                                          [(equal? name 'R-PAR) ")"]
                                          [(equal? name 'COMMA) ","]
                                          [(equal? name 'SEMI-COLON) ";"]
                                          [(equal? name 'MINUS) "-"]
                                          [(equal? name 'PLUS) "+"]
                                          [(equal? name 'DOT) "."]
                                          [(equal? name 'ASSIGN) "="]
                                          [(equal? name 'EQ) "=="]
                                          [(equal? name 'GREATER) ">"]
                                          [(equal? name 'GREATER-EQ) ">="]
                                          [(equal? name 'LOWER) "<"]
                                          [(equal? name 'LOWER-EQ) "<="]
                                          [(equal? name 'NOT-EQ) "!="]
                                          [(equal? name 'R-ARROW) "->"]
                                          [else (symbol->string name)])]
					[(string? value) value]
					[(number? value) (number->string value)]
					[(symbol? value) (symbol->string value)]
					[else (first value)]
					)
                                      "' al line " (number->string (counter 'get))))))

		   (grammar

		     (query 
		       [(statement-list)  #t])

		     (statement-list 
		       [(statement) #f]
		       [(statement statement-list) #f])

		     (statement
		       [(select-statement SEMI-COLON) #f]
		       [(project-statement SEMI-COLON) #f]
		       [(map-statement SEMI-COLON) #f]
		       [(join-statement SEMI-COLON) #f]
		       [(materialize-statement SEMI-COLON) #f]
		       [(union-statement SEMI-COLON) #f]
		       [(region-intersection-statement SEMI-COLON) #f]
		       [(difference-statement SEMI-COLON) #f]
		       [(aggregate-statement SEMI-COLON) #f]
                       [(order-statement SEMI-COLON) #f])
                                                                 
		     (select-statement 
		       [(dataset-identifier ASSIGN SELECT L-PAR select-condition  R-PAR dataset-identifier) 
			(letrec ((left $1)
				 (right $7)
				 (condition $5)
				 (translated (parser-utils:emit-select-code global-schemata left condition right)))
			  (begin
			    (set! preamble (string-append preamble (first translated) ))
			    (set! pig-code (string-append pig-code (second translated) ))
			    ))]
		       [(dataset-identifier ASSIGN SELECT L-PAR select-all-symbol R-PAR dataset-identifier) 
			(letrec
			  ((left $1)
			   (right $7)
			   (translated (parser-utils:emit-select-all-code global-schemata left right))) 
			  (begin
			    (set! preamble (string-append preamble (first translated)))
			    (set! pig-code (string-append pig-code (second translated) ))))])
                     
                     (select-all-symbol
                      [(TRUE) #f]
                      [(STAR) #f])

		     (select-condition
		       [(select-expr) $1])
		     (select-expr
		       [(select-term) $1]
		       [(select-term OR select-expr) (append $1 (list "+") $3)])
		     (select-term 
		       [(select-fact) $1]
		       [(select-fact AND select-term) (append $1 (list "*") $3)])

		     (select-fact
		       [(select-check) $1]
		       [(L-PAR select-expr R-PAR) (append (list "(") $2 (list ")"))]
		       [(NOT L-PAR select-expr R-PAR) (append (list "{") $3 (list "}"))])
		     (select-check
		       [(field-identifier select-check-operator select-check-value) (let 
										      ((path (if (list? (field-instance-path $1))
											       (field-instance-path $1)
											       (list)))
										       (name (field-instance-identifier $1)))
										      (append (list "[")
											      (list (foldr (λ (x y) (string-append x "." y)) name (if (list? path) path (list))))
											      $2 
											      $3 
											      (list "]")))])
		     (select-check-operator
		       [(EQ) (list "=")]
		       [(NOT-EQ) (list "!=")])

		     (select-check-value
		       [(INT-VALUE) (list  (number->string $1) )]
		       [(FLOAT-VALUE) (list (number->string $1))]
		       [(STRING-VALUE) (list $1)])

		     (project-statement
		       [(dataset-identifier ASSIGN PROJECT L-PAR project-condition tuple-modifier R-PAR dataset-identifier) 
			(letrec
			  ((left-var $1)
			   (condition $5)
			   (region-modifier (first $6))
                           (attribute-modifier (second $6))
                           (right-var $8)
			   (translated (parser-utils:emit-project-code global-schemata left-var condition region-modifier attribute-modifier right-var)))
                          (set! preamble (string-append preamble (first translated)))
			  (set! pig-code (string-append pig-code (second translated))))])

		     (project-condition
                      [(TRUE) 'TRUE]
                      [(STAR) 'TRUE]
                      [(project-expr)  $1])
		     (project-expr
                      [(project-term) $1]
                      [(project-term OR project-expr) (list "or" $1 $3)])
		     (project-term
                      [(project-fact) $1]
                      [(project-fact AND project-term) (list "and" $1 $3)])
		     (project-fact
                      [(project-check) $1]
                      [(L-PAR project-expr R-PAR) (list $2)])
		     (project-check
                      [(project-field-value-check) $1]
                      [(project-field-field-check) $1]
                      [(project-region-check) $1])

		     (project-field-value-check
		       [(field-identifier project-check-operator project-check-value) 
			(let ((operator $2)
			      (field $1)
			      (value $3))
			  (list operator field value))])


		     (project-region-check
		       [(LEFT project-check-operator INT-VALUE) (list $2 'left $3)]
		       [(RIGHT project-check-operator INT-VALUE) (list $2 'right $3)]
		       [(STRAND EQ strand-value) (list "==" 'strand $3)]
		       [(CHR EQ STRING-VALUE) (list "==" 'chr (if
                                                               (equal? "chr" (string-downcase (substring $3 0 3)))
                                                               (substring $3 3)
                                                               $3))])

		     (strand-value
		       [(STAR) "'*'"]
		       [(PLUS) "'+'"]
		       [(MINUS) "'-'"])

		     (project-check-operator
		       [(EQ) "=="]
		       [(NOT-EQ) "!="]
		       [(GREATER) ">"]
		       [(LOWER) "<"]
		       [(LOWER-EQ) "<="]
		       [(GREATER-EQ) ">="])

		     (project-check-value
		       [(float-value) $1]
		       [(integer-value) $1]
		       [(STRING-VALUE) $1])

		     (integer-value
		       [(INT-VALUE) $1]
		       [(MINUS INT-VALUE) (- $2)])

		     (float-value
		       [(FLOAT-VALUE) $1]
		       [(MINUS FLOAT-VALUE) (- $2)])

		     (project-field-field-check
		       [(field-identifier project-check-operator field-identifier) 
			(let ((operator $2)
			      (left-op $1)
			      (right-op $3))
			  (list operator left-op right-op))])
                     
                     (tuple-modifier
                      [() (list (list) (list))]
                      [(SEMI-COLON attribute-modifier) (list (list) $2)]
                      [(SEMI-COLON region-modifier) (list $2 (list))]
                      [(SEMI-COLON region-modifier COMMA attribute-modifier) (list $2 $4)])
                     
                     (attribute-modifier
                      [(single-attribute-modifier) (list $1)]
                      [(single-attribute-modifier COMMA attribute-modifier) (append (list $1) $3)])
                     
                     (single-attribute-modifier
                      [(string-constant) $1]
                      [(algebraic-modifier) $1])
                     
                     (algebraic-modifier
                      [(field-identifier AS mod-exp) (list $1 $3)])
                      
                     
                     (mod-exp
                      [(mod-term) $1]
                      [(mod-term PLUS mod-exp) (list "+" $1 $3)]
                      [(mod-term MINUS mod-exp) (list "-" $1 $3)])
                     
                     (mod-term
                      [(mod-factor) $1]
                      [(mod-factor STAR mod-term) (list "*" $1 $3)]
                      [(mod-factor DIV mod-term) (list "/" $1 $3)])
                     
                     (mod-factor
                      [(INT-VALUE) (number->string $1)]
                      [(FLOAT-VALUE) (number->string $1)]
                      [(LEFT) 'LEFT]
                      [(RIGHT) 'RIGHT]
                      [(field-identifier) $1]
                      [(MINUS mod-factor) (list "-" "0" $2)]
                      [(L-PAR mod-exp R-PAR) (list $2)])
                     
                     (string-constant
                      [(field-identifier AS STRING-VALUE) (list $1 (string-append "'" $3 "'"))])

		     (region-modifier
                      [(start-modifier COMMA stop-modifier )  (list $1 $3)]
                      [(stop-modifier COMMA start-modifier) (list $3 $1)]
                      [(LEFT ASSIGN field-identifier) (list 'FROM-FIELD (list 'LEFT $3) (list))]
                      [(RIGHT ASSIGN field-identifier) (list 'FROM-FIELD (list) (list 'RIGHT $3))]
                      [(RIGHT ASSIGN field-identifier AND LEFT ASSIGN field-identifier) (list 
                                                                                            'FROM-FIELD
                                                                                            (list 'LEFT $7)
                                                                                            (list 'RIGHT $3))]
                      [(LEFT ASSIGN field-identifier AND RIGHT ASSIGN field-identifier) (list 
                                                                                           'FROM-FIELD
                                                                                           (list 'LEFT $3)
                                                                                           (list 'RIGHT $7))])
                     (start-modifier
		       [(START ASSIGN region-end-identifier region-modifier-operator INT-VALUE) (list $3 $4 $5)]
		       [(START ASSIGN region-end-identifier) (list $3 "+" 0)])

		     (stop-modifier
		       [(STOP ASSIGN region-end-identifier region-modifier-operator INT-VALUE) (list $3 $4 $5)]
		       [(STOP ASSIGN region-end-identifier) (list $3 "+" 0)])

		     (region-end-identifier
		       [(fixed-identifier) $1]
		       [(relative-identifier) $1])

		     (fixed-identifier
		       [(LEFT) "left"]
		       [(RIGHT) "right"])

		     (relative-identifier
		       [(START) "start"]
		       [(STOP) "stop"])

		     (region-modifier-operator
		       [(PLUS) "+"]
		       [(MINUS) "-"])


		     (map-statement
		       [(dataset-identifier ASSIGN map-type L-PAR map-meta-filter map-function-list R-PAR reference dataset ) 
			(letrec
			  ((left $1)
			   (meta-cond $5)
			   (aggregate-function $6)
			   (experiment-regions $9)
			   (referiment-regions $8)
			   (type $3)
			   (translated (parser-utils:emit-map-code global-schemata left type meta-cond aggregate-function referiment-regions experiment-regions)))
			  (set! preamble (string-append preamble (first translated)))
			  (set! pig-code (string-append pig-code (second translated))))])

		     (map-type
		       [(MAP) "not-stranded"]
		       [(MAP-STRANDED) "stranded"])

		     (map-meta-filter
		       [() 'none]
		       [(join-metadata-condition COMMA) $1])

		     (map-function-list
		       [(map-function) (list $1)]
		       [(map-function-list COMMA map-function) (append $1 (list $3))])

		     (map-function
		       [(BUILTIN-NULLARY-FUNCTION-NAME) (list (string-downcase $1) $1)]
                       [(field-identifier AS BUILTIN-NULLARY-FUNCTION-NAME) (list (field-instance-identifier $1) $3)]
		       [(BUILTIN-UNARY-FUNCTION-NAME L-PAR field-identifier R-PAR) (list (string-downcase $1) $1 $3)]
                       [(field-identifier AS BUILTIN-UNARY-FUNCTION-NAME L-PAR field-identifier R-PAR) (list (field-instance-identifier $1) $3 $5)])

		     (join-statement
		       [(dataset-identifier ASSIGN join-type L-PAR join-arguments COMMA join-composition-function R-PAR left-input right-input ) 
			(letrec
			  ((type $3)
			   (left $1)
			   (genometric-condition (first $5))
			   (metadata-condition (if (> (length $5) 1) (second $5) 'not-set))
			   (composition $7)
			   (input-left $9)
			   (input-right $10)
			   (translated 
			     (parser-utils:emit-join-code 
			       global-schemata left metadata-condition genometric-condition composition type input-left input-right)))
			  (set! preamble (string-append preamble (first translated)))
			  (set! pig-code (string-append pig-code (second translated))))])

		     (join-type
		       [(JOIN) "not-stranded"]
		       [(JOIN-STRANDED) "stranded"])

		     (right-input
		       [(dataset-identifier) $1])

		     (left-input
		       [(dataset-identifier) $1])

		     (join-arguments
		       [(join-region-condition ) (list $1)]
		       [(join-metadata-condition COMMA join-region-condition) (list $3 $1)])

		     (join-composition-function
		       [(INTERSECTION) "and"]
		       [(CONJUNCTION) "con"]
		       [(LEFT) "r1"]
		       [(RIGHT) "r2"]
		       [(LEFT-DISTINCT) "r1_distinct"]
		       [(RIGHT-DISTINCT) "r2_distinct"]
		       [(PROJECT-LEFT) "project_r1"]
		       [(PROJECT-RIGHT) "project_r2"]
		       [(PROJECT-LEFT-DISTINCT) "project_r1_distinct"]
		       [(PROJECT-RIGHT-DISTINCT) "project_r2_distinct"])


		     (join-region-condition
		       [(simple-distance-join-condition) $1]
		       [(complex-distance-join-condition) $1]
		       [(simple-mindistance-join-condition) $1]
		       [(complex-mindistance-join-condition) $1]
		       [(simple-first-after-join-condition) $1]
		       [(complex-first-after-join-condition) $1])

		     (simple-distance-join-condition
		       [(DISTANCE LOWER INT-VALUE) (list "distanceLessThan" (number->string $3) (number->string $3))]
		       [(UPSTREAM-DISTANCE LOWER INT-VALUE) (list "distanceLessThan" (number->string $3) "NONE")]
		       [(DOWNSTREAM-DISTANCE LOWER INT-VALUE) (list "distanceLessThan" "NONE" (number->string $3))]
		       [(UPSTREAM-DISTANCE LOWER INT-VALUE OR DOWNSTREAM-DISTANCE LOWER INT-VALUE) (list
												     "distanceLessThan" 
												     (number->string $3)
												     (number->string $7))]
		       [(DOWNSTREAM-DISTANCE LOWER INT-VALUE OR UPSTREAM-DISTANCE LOWER INT-VALUE) (list
												     "distanceLessThan"
												     (number->string $7)
												     (number->string $3))])

		     (complex-distance-join-condition
		       [(DISTANCE LOWER INT-VALUE AND DISTANCE GREATER INT-VALUE) (list 
										    "distanceLessThan" 
										    (number->string $3) 
										    (number->string $3)
										    "distanceGreaterThan"
										    (number->string $7)
										    (number->string $7))]
		       [(UPSTREAM-DISTANCE LOWER INT-VALUE AND UPSTREAM-DISTANCE GREATER INT-VALUE) (list
												      "distanceLessThan" 
												      (number->string $3)
												      "NONE"
												      "distanceGreaterThan"
												      (number->string $7)
												      "NONE"
												      )]
		       [(DOWNSTREAM-DISTANCE LOWER INT-VALUE AND DOWNSTREAM-DISTANCE GREATER INT-VALUE) (list
													  "distanceLessThan"
													  "NONE"
													  (number->string $3)
													  "distanceGreaterThan"
													  "NONE"
													  (number->string $7))])
		     (simple-mindistance-join-condition
		       [(MIN-DISTANCE L-PAR INT-VALUE R-PAR) (list "minDistance" (number->string $3) (number->string $3))])

		     (complex-mindistance-join-condition
		       [(MIN-DISTANCE L-PAR INT-VALUE R-PAR AND DISTANCE LOWER INT-VALUE) 
			(list "minDistance" 
			      (number->string $3) 
			      (number->string $3)
			      "distanceLessThan"
			      (number->string $8)
			      (number->string $8))]
		       [(MIN-DISTANCE L-PAR INT-VALUE R-PAR AND DISTANCE GREATER INT-VALUE) 
			(list "minDistance" 
			      (number->string $3) 
			      (number->string $3)
			      "distanceGreaterThan"
			      (number->string $8)
			      (number->string $8))]
		       )

		     (simple-first-after-join-condition
		       [(FIRST L-PAR INT-VALUE R-PAR AFTER DISTANCE INT-VALUE) (list 
										 "firstAfter" 
										 (number->string $3) 
										 (number->string $7))]
		       [(FIRST L-PAR INT-VALUE R-PAR AFTER UPSTREAM-DISTANCE INT-VALUE) (list 
											  "firstAfterUpstream" 
											  (number->string $3) 
											  (number->string $7))]
		       [(FIRST L-PAR INT-VALUE R-PAR AFTER DOWNSTREAM-DISTANCE INT-VALUE) (list 
											    "firstAfterDownstream" 
											    (number->string $3) 
											    (number->string $7))])


		     (complex-first-after-join-condition
		       [(FIRST L-PAR INT-VALUE R-PAR AFTER DISTANCE INT-VALUE AND DISTANCE LOWER INT-VALUE) 
			(list 
			  "firstAfter" 
			  (number->string $3) 
			  (number->string $7)
			  "distanceLessThan"
			  (number->string $11)
			  (number->string $11))]
		       [(FIRST L-PAR INT-VALUE R-PAR AFTER UPSTREAM-DISTANCE INT-VALUE AND UPSTREAM-DISTANCE LOWER INT-VALUE) 
			(list 
			  "firstAfterUpstream" 
			  (number->string $3) 
			  (number->string $7)
			  "distanceLessThan"
			  (number->string $11)
			  "NONE")]
		       [(FIRST L-PAR INT-VALUE R-PAR AFTER DOWNSTREAM-DISTANCE INT-VALUE AND DOWNSTREAM-DISTANCE LOWER INT-VALUE) 
			(list 
			  "firstAfterDownstream" 
			  (number->string $3) 
			  (number->string $7)
			  "distanceLessThan"
			  "NONE"
			  (number->string $11))])

		     (join-metadata-condition
		       [(join-metadata-expr) $1])

		     (join-metadata-expr
		       [(join-metadata-term) $1]
		       [(join-metadata-term OR join-metadata-expr) (string-append $1 "+" $3)])

		     (join-metadata-term
		       [(join-metadata-fact) $1]
		       [(join-metadata-fact AND join-metadata-term) (string-append $1 "*" $3)])

		     (join-metadata-fact 
		       [(join-metadata-check) $1]
		       [(L-PAR join-metadata-expr R-PAR) (string-append "(" $2 ")")])

		     (join-metadata-check
		       [(join-metadata-single-check) $1]
		       [(join-metadata-double-check) $1])

		     (join-metadata-single-check
		       [(side R-ARROW field-identifier EQ join-metadata-single-check-value) 
			(string-append "[" $1 "#" (field-instance->identifier $3) "=" $5 "]")])

		     (join-metadata-double-check
		       [(side R-ARROW field-identifier EQ side R-ARROW field-identifier) 
			(string-append "[" $1 "#" (field-instance->identifier $3) "=" $5 "#" (field-instance->identifier $7) "]") ])


		     (side
		       [(RIGHT) "right"]
		       [(LEFT) "left"])

		     (join-metadata-single-check-value
		       [(STRING-VALUE) $1]
		       [(INT-VALUE) (number->string $1)]
		       [(FLOAT-VALUE) (number->string $1)])

		     (union-statement
		       [(dataset-identifier ASSIGN UNION dataset-identifier dataset-identifier-list) 
			(letrec
			  ((new-var $1)
			   (operand-list (append (list $4) $5))
			   (translated (parser-utils:emit-union-code global-schemata new-var operand-list )))
			  (begin
			    (set! preamble (string-append preamble (first translated)))
			    (set! pig-code (string-append pig-code (second translated)))))])

		     (dataset-identifier-list
		       [(dataset-identifier) (list $1)]
		       [(dataset-identifier-list dataset-identifier) (append $1 (list $2))])

		     (difference-statement
		       [(dataset-identifier ASSIGN difference-type L-PAR difference-meta-data-filter R-PAR dataset-identifier dataset-identifier)
			(letrec
			  ((new-var $1)
			   (type $3)
			   (meta-filter $5)
			   (to-be-filtered $7)
			   (filter-regions $8)
			   (translated (parser-utils:emit-difference-code global-schemata new-var type meta-filter to-be-filtered filter-regions)))
			  (begin 
			    (set! preamble (string-append preamble (first translated)))
			    (set! pig-code (string-append pig-code (second translated)))))])

		     (difference-type 
		       [(DIFFERENCE) "not-stranded"]
		       [(DIFFERENCE-STRANDED) "stranded"])

		     (difference-meta-data-filter
		       [() 'none]
		       [(join-metadata-condition) $1])


		     (region-intersection-statement
		       [(dataset-identifier ASSIGN intersection-operation-type L-PAR min-acc COMMA max-acc intersection-optional-par R-PAR cover-input-datasets) 
			(letrec
			  ((left-var $1)
			   (type $3)
			   (minimum-acc $5)
			   (maximum-acc $7)
			   (jacc-min (first (first $8)))
			   (jacc-max (second (first $8)))
			   (group-var (second $8))
			   (merge (third $8))
                           (func-list (fourth $8))
			   (datasets $10)
			   (translated (parser-utils:emit-region-intersection-code global-schemata left-var type minimum-acc maximum-acc jacc-min jacc-max group-var merge datasets func-list)))
			  (begin
			    (set! preamble (string-append preamble (first translated)))
			    (set! pig-code (string-append pig-code (second translated)))))])

                     (cover-aggregate-functions
                      [(SEMI-COLON cover-function-list) $2])
                     
                     (cover-function-list
                      [(cover-function-call) (list $1)]
                      [(cover-function-call COMMA cover-function-list) (append (list $1) $3)])
                     
                     (cover-function-call
                      [(BUILTIN-NULLARY-FUNCTION-NAME) (list $1 (field-instance (string-downcase $1) (counter 'get) 'not-defined))]
                      [(field-identifier AS BUILTIN-NULLARY-FUNCTION-NAME) (list $3 $1)]
                      [(BUILTIN-UNARY-FUNCTION-NAME L-PAR field-identifier R-PAR)  (list $1 (field-instance (string-downcase $1) (counter 'get) 'not-defined) $3)]
                      [(field-identifier AS BUILTIN-UNARY-FUNCTION-NAME L-PAR field-identifier R-PAR) (list $3 $1 $5)])
                     
                      
                     (intersection-operation-type
                      [(COVER) 'cover]
                      [(COVER-FLAT) 'cover-flat]
                      [(SUMMIT) 'summit]
                      [(SUMMIT-FLAT) 'summit-flat])
                    

		     (cover-input-datasets
		       [(dataset-identifier) (list $1)]
		       [(dataset-identifier dataset-identifier) (list $1 $2)])


		     (min-acc
		       [(ALL) "all"]
		       [(ALL MINUS INT-VALUE) (string-append "all" "-" (number->string $3))]
		       [(ALL PLUS INT-VALUE) (string-append "all" "+" (number->string $3))]
		       [(INT-VALUE) (number->string $1)])

		     (max-acc
		       [(ALL) "all"]
		       [(ALL MINUS INT-VALUE) (string-append "all" "-" (number->string $3))]
		       [(ALL PLUS INT-VALUE) (string-append "all" "+" (number->string $3))]
		       [(ANY) "any"]
		       [(INT-VALUE) (number->string $1)])

		     (intersection-optional-par
                      [() (list (list "none" "none") (list) #f (list))]
		       [(jaccard-options) (list $1 (list) #f (list))]
		       [(jaccard-options group-by-option) (list $1 $2 #f (list))]
		       [(jaccard-options merge-option) (list $1 (list) #f (list))]
		       [(group-by-option) (list (list "none" "none") $1 #f (list))]
		       [(group-by-option merge-option) (list (list "none" "none") $1 $2 (list))]
		       [(merge-option) (list (list "none" "none") (list) $1 (list))]
		       [(jaccard-options group-by-option merge-option) (list $1 $2 $3 (list))]
		       [(cover-aggregate-functions) (list (list "none" "none") (list) #f $1)]
		       [(jaccard-options cover-aggregate-functions) (list $1 (list) #f $2)]
		       [(jaccard-options group-by-option cover-aggregate-functions) (list $1 $2 #f $3)]
		       [(jaccard-options merge-option cover-aggregate-functions) (list $1 (list) #f $3)]
		       [(group-by-option cover-aggregate-functions) (list (list "none" "none") $1 #f $2)]
		       [(group-by-option merge-option cover-aggregate-functions) (list (list "none" "none") $1 $2 $3)]
		       [(merge-option cover-aggregate-functions) (list (list "none" "none") (list) $1 $2)]
		       [(jaccard-options group-by-option merge-option cover-aggregate-functions) (list $1 $2 $3 $4)])


		     (jaccard-options
		       [(SEMI-COLON JACCARD FLOAT-VALUE COMMA FLOAT-VALUE) 
			(if
			  (and (>= $3 0)
			       (<= $3 1)
			       (>= $5 0)
			       (<= $5 1)
			       (<= $3 $5))
			  (list (number->string $3) (number->string $5))
			  (error "Jaccard-min and Jaccard-max have to be in [0,1]. Moreover, Jaccard-max must be greater or equal than Jaccard-min"))])


		     (group-by-option
		       [(SEMI-COLON GROUP BY groupby-field-list) $4])

		     (groupby-field-list
		       [(field-identifier) (list $1)]
		       [(field-identifier COMMA groupby-field-list) (append (list $1) $3)])

		     (merge-option
		       [(SEMI-COLON MERGE) #t])

		     (materialize-statement
		       [(MATERIALIZE dataset-identifier) 
			(letrec
			  ((var $2)
			   (var-name (variable-identifier var))
			   (var-owner (variable-user var))
			   (translated (parser-utils:emit-materialize-code global-schemata var)))
			  (begin
			    (set! preamble (string-append preamble (first translated) ))
			    (set! pig-code (string-append pig-code (second translated) ))
			    (set! output-schemata (string-append output-schemata (global-schemata 'get-xml var-name var-owner)))))])

		     (aggregate-statement
		       [(dataset-identifier ASSIGN AGGREGATE L-PAR aggregate-function-list R-PAR dataset-identifier) 
			(letrec
			  ((aggregate-function-list $5)
			   (new-var $1)
			   (input-var $7)
			   (translated (parser-utils:emit-aggregate-code global-schemata new-var aggregate-function-list input-var)))
			  (begin
			    (set! preamble (string-append preamble (first translated)))
			    (set! pig-code (string-append pig-code (second translated)))))])

		     ;each function token generates a list:
		     ;nullary: (list function new-name)
		     ;unary: (list function new-name argument)
		     (aggregate-function-list
		       [(single-aggregate-function) $1]
		       [(aggregate-function-list COMMA single-aggregate-function) (append $1 $3)])

		     (single-aggregate-function
		       [(nullary-aggregate-function) (list $1)]
		       [(unary-aggregate-function) (list $1)])

		     (nullary-aggregate-function
		       [(BUILTIN-NULLARY-FUNCTION-NAME) (list $1 (field-instance (string-downcase $1) (counter 'get) 'not-defined)) ]
		       [(field-identifier AS BUILTIN-NULLARY-FUNCTION-NAME) (list $3 $1)])

		     (unary-aggregate-function
		       [(BUILTIN-UNARY-FUNCTION-NAME L-PAR field-identifier R-PAR) (list $1 (field-instance (string-downcase $1) (counter 'get) 'not-defined) $3)]
		       [(field-identifier AS BUILTIN-UNARY-FUNCTION-NAME L-PAR field-identifier R-PAR) (list $3 $1 $5)])

                     
                     (order-statement
                      [(dataset-identifier ASSIGN ORDER L-PAR order-arguments R-PAR dataset-identifier) 
                       (letrec
                           ((left-var $1)
                            (right-var $7)
                            (ordering (first $5))
                            (top (second $5))
                            (translated (parser-utils:emit-order-code global-schemata left-var right-var ordering top)))
                         (begin
                            (set! preamble (string-append preamble (first translated)))
			    (set! pig-code (string-append pig-code (second translated)))))])

                     
                     (order-arguments
                      [(order-clauses) (list $1 (list))]
                      [(order-clauses SEMI-COLON top-option) (list $1 $3)])
                     
                     (order-clauses
                      [(order-clause) (list $1)]
                      [(order-clause COMMA order-clauses) (cons $1 $3)])
                     
                     (order-clause
                      [(field-identifier) (list 'ASC $1)]
                      [(DESC field-identifier) (list 'DESC $2)])
                     
                     (top-option
                      [(TOP INT-VALUE) (list 'TOP $2)]
                      [(TOPG INT-VALUE) (list 'TOPG $2)])
                     
                     (dataset
		       [(dataset-identifier) $1])
		     (reference
		       [(dataset-identifier) $1])

		     (dataset-identifier
		       [(VAR-ID-VALUE) (let 
					 ((identifier (first $1))
					  (line-number (second $1))) 
					 (variable identifier line-number 'none))]
		       [(VAR-ID-VALUE DOT VAR-ID-VALUE) (let
							  ((identifier (first $3))
							   (line-number (second $3))
							   (user (first $1)))
							  (variable identifier line-number user))]) 
		     (field-identifier
		       [(VAR-ID-VALUE) (let
					 ((identifier (first $1))
					  (line-number (second $1))
					  (path 'not-defined))
					 (field-instance identifier line-number path))]
		       [(field-identifier-path DOT VAR-ID-VALUE) (let
								   ((identifier (first $3))
								    (line-number (second $3))
								    (path $1))
								   (field-instance identifier line-number path))])

		     (field-identifier-path
		       [(VAR-ID-VALUE) (cons (first $1) '())]
		       [(field-identifier-path DOT VAR-ID-VALUE) (append $1 (list (first $3)))])
		     )

		   ))

     (return-output-xml-schema
       (λ () 
	  (string-append
	    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
	    "<gmqlSchemaCollection name=\"GLOBAL_SCHEMAS\" xmlns=\"http://genomic.elet.polimi.it/entities\">\n"
	    output-schemata
	    "</gmqlSchemaCollection>")))


     )
    (λ args
       (let
	 ((cmd (first args)))
	 (cond
	   ((equal? cmd 'parse-input-port) (let ((port (second args))) 
					     (begin
					       (let
						 ((translated (parser-utils:emit-pre-parsing-code global-schemata)))
						 (begin
						   (set! pig-code (string-append pig-code (second translated)))
						   (set! preamble (string-append preamble (first translated)))))
					       (gql-parser (λ () (lex-token port)))
					       (let
						 ((translated (parser-utils:emit-post-parsing-code global-schemata)))
						 (begin
						   (set! pig-code (string-append pig-code (second translated)))
						   (set! preamble (string-append preamble (first translated))))))))
	   ((equal? cmd 'get-preamble) preamble)
	   ((equal? cmd 'get-pig-code) pig-code)
	   ((equal? cmd 'get-schemata) (return-output-xml-schema)) 
	   (else (error (string-append "parser: unknow command " (symbol->string cmd)))))))))

