#lang racket


(require parser-tools/yacc
	 parser-tools/lex
	 (prefix-in : parser-tools/lex-sre))

(provide remove-comments)

(define-empty-tokens
  op-tokens
  (COMMENT
   NEW-LINE
   EOF))

(define-tokens value-tokens
	       (OTHER))

(define-lex-abbrevs
  [COMMENT (:seq "#" (:* (:/ "0" "9") (:/ "a" "z") (:/ "A" "Z") 
			 "." ":" "_" "+" "*" 
			 "-" "/" "(" ")" "," 
			 ";" "[" "]" "<" ">" 
			 "=" "!" "'" " " "\t"
			 "è" "é" "ì" "ò" "à"
			 "ù" "?" "|" "&" "{" 
			 "}" "\"") )]
  [OTHER (:seq (:+ (:/ "0" "9") (:/ "a" "z") (:/ "A" "Z") 
		   "." ":" "_" "+" "*" 
		   "-" "/" "(" ")" "," 
		   ";" "[" "]" "<" ">" 
		   "=" "!" "'" " " "\t") )])

(define lex-token (lexer
		    [(eof) 'EOF]
		    [(:seq #\newline) 'NEW-LINE]
		    [(:seq COMMENT) 'COMMENT]
		    [(:seq OTHER) (token-OTHER lexeme) ]))

(define comments-parser (parser
			  (start query)
			  (end EOF)
			  (tokens op-tokens value-tokens)
			  (suppress)
			  (error (λ(ok? name value) (error "FATAL: Preprocessor error")))

			  (grammar
			    (query
			      [(list-of-lines) $1])
			    
			    (list-of-lines
			      [(line) $1]
			      [(line list-of-lines) (string-append $1 $2)])

			    (line
			      [(COMMENT NEW-LINE) "\n"]
			      [(OTHER NEW-LINE) (string-append $1 "\n")]
			      [(OTHER COMMENT NEW-LINE) (string-append $1 "\n")]
			      [(NEW-LINE) "\n"]
			      [(COMMENT) "\n"]
			      [(OTHER) $1]
			      [(OTHER COMMENT) $1]))))

;remove all the comments
;i.e. substitute "#{any}*\n" with "\n"
;x
(define (remove-comments x)
  (open-input-string (comments-parser (λ () (lex-token x)))))
