#lang racket

(provide input-file-path 
	 generated-pig-code-file-path
	 allowed-users
	 materialize-directory
	 data-repository-root
	 generated-schema-output-file
	 region-prefix)

(require racket/cmdline)
(require "configurations.rkt")

(define generated-pig-code-file-path default-generated-pig-code-file-path)
(define generated-schema-output-file (string-append generated-pig-code-file-path ".schema"))
(define input-file-path 'null)
(define materialize-directory default-hadoop-materialize-directory)
(define data-repository-root hadoop-repository-root)
(define region-prefix hadoop-region-repository-prefix)

;initializes the list of user: public is by default always present.
(define allowed-users default-user-list)
(define (add-allowed-user u)
  (if (member u allowed-users)
    (displayln (string-append "User " u " is already in the list. It won't be added again."))
    (set! allowed-users (cons u allowed-users))))

(define :user "public")
(define :job-id 'null)


(command-line
  #:program "GenoMetric Query Language Compiler"
  #:once-each
  [("-o" "--output-file") output-file "Specifies the full path where the PIG generated code will be written. [default: \"./a.out\"]" 
			 (begin
			   (set! generated-pig-code-file-path output-file)
			   (set! generated-schema-output-file (string-append generated-pig-code-file-path ".schema")))]
  [("-s" "--script-id") ID "Specifies the ID of the script. (This is used for creating the output directory)."
			(set! :job-id ID)]  
  [("-l") "Enable local exectution of the query"
	  (begin
	    (set! materialize-directory default-local-materialize-directory)
	    (set! data-repository-root local-repository-root)
	    (set! region-prefix local-region-repository-prefix))]
  
  #:multi
  [("-u" "--user") user "This option can be specified multiple times; all the datasets owned by 'user' will be added to the available ones."
		   (begin	      
		     (set! :user user)
		     (add-allowed-user user))]

  #:args (filename) 
  (begin
    (set! materialize-directory (string-append materialize-directory "/" :user "/results/" :job-id "/")) 
    (set! input-file-path filename)))
