#lang racket

(require xml)
(require racket/match)

(provide 
  trim-list
  get-attribute-value
  get-optional-attribute-value
  file->xexpr
  xml->file-list)

;remove from a list all the strings that are made of only [:space:] characters
(define (trim-list x)
  (if (null? x) 
      '()
      (match (car x)
        [(regexp #px"[[:space:]]+") (trim-list (cdr x))]
        [_ (cons (car x) (trim-list (cdr x)))])))


;given a list of attributes of a XML elements, returns the value
;associated with the specified name
(define (get-attribute-value attribute-name attribute-list)
  (cond
    [(null? attribute-list) (error (string-append "Field " attribute-name + " not found"))]
    [(equal? (car (car attribute-list)) attribute-name) (second (car attribute-list))]
    [else (get-attribute-value attribute-name (cdr attribute-list))]))

;given a list of attributes either return the value associated with a given name
; or the default value (if the attribute is not specified)
(define (get-optional-attribute-value attribute-name default-value attribute-list)
  (cond
    [(null? attribute-list) default-value]
    [(equal? (car (car attribute-list)) attribute-name) (second (car attribute-list))]
    [else (get-attribute-value attribute-name (cdr attribute-list))]))

;open a file and return a xexpr
(define (file->xexpr url)
 (xml->xexpr (document-element (read-xml (open-input-file url)))))

;given a dataset file, return the list of all the dataset
;xml : is an xexpr
(define (xml->file-list xml)
  (let ((x (trim-list xml)))
    (match x
	   [(list 'DataSets attributes dataset-list) (xml->file-list dataset-list)]
	   [(list 'dataset attributes dataset-list ...) (map (λ(x) (xml->file-list (trim-list x))) dataset-list)]
	   [(list 'url attributes link) link])))
